// ==============================================================
//                 ORBITER MODULE: bc_tower
//
// bc_tower.cpp
// Control module for bc_tower vessel class
//
// ==============================================================

#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "gcAPI.h"
#include <cmath>

// ==============================================================
// Some vessel parameters
// ==============================================================
const DWORD COMM_SEND_CT = 2399;

const double ARM_SLEW_SPEED = 0.125;
const double ARM_GRAB_SPEED = 0.2;
const double ARM_TRANS_SPEED = 0.1;
const double QDA_DETACH_SPEED = 0.2;
const double QDA_ATTACH_SPEED = 0.1;
const double CLAW_SPEED = 0.2;

double arm_trans_length = 100;

VECTOR3 bsred = _V(1,0.1,0.1);

VECTOR3 vA = _V(5.098,0,-4.651);
VECTOR3 vB = _V(1.223,0,-7.571);
VECTOR3 vC = _V(1.984,0,-10.782);
double claw_trans_length = -6.7;

VECTOR3 piston_hinge = _V(1.347,0,-40.435);
VECTOR3 piston_end = _V(9.394,0,-40.435);
VECTOR3 piston_attach = _V(9.034,0,-39.796);
VECTOR3 qda_rot = _V(6.52,0,-36.883);
double piston_trans_length = -5;

OBJHANDLE hGref;


static const DWORD ntdvtx_geardown = 3;
static TOUCHDOWNVTX tdvtx_geardown[ntdvtx_geardown] = {
	{_V(    0,-0.5,50), 1e6, 1e6, 0.7, 0.7},
	{_V(-43.3,-0.5,-25), 1e6, 1e6, 0.7, 0.7},
	{_V( 43.3,-0.5,-25), 1e6, 1e6, 0.7, 0.7}
};


class bc_tower: public VESSEL3 {
public:
	bc_tower (OBJHANDLE hVessel, int flightmodel);
	~bc_tower ();
	void DefineAnimations (void);
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep(double SimT, double SimDT, double MJD);
	bool clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp);
	void clbkPostCreation();
	void clbkSaveState (FILEHANDLE scn);
	void clbkLoadStateEx (FILEHANDLE scn, void *status);
	int clbkConsumeBufferedKey(DWORD key, bool down, char *kstate);
	void clbkVisualCreated (VISHANDLE vis, int refcount);

private:
	void DoCatch(void);
	void EnableCatch(void);
	void DisableCatch(void);
	double step_time;
	UINT anim_arm_slew, anim_arm_grab, anim_arm_trans, anim_qda, anim_cables, anim_box;
	UINT anim_claw, anim_clawrot1, anim_clawrot2;
	UINT anim_piston_rot, anim_piston_trans;
	MGROUP_TRANSFORM *ArmSlew, *ArmGrab1,*ArmGrab2, *ArmTrans, *ArmCables, *QDARot, *QDATrans, *QDABoxRot, *QDABoxTrans;
	MGROUP_TRANSFORM *ClawRetract, *ClawRot1, *ClawRot2, *ClawRot3, *ClawRot4;
	MGROUP_TRANSFORM *PistonRot, *PistonTrans;
	int light_status, lox_status, launch_status, catch_status, caught;
	int arm_slew_status, arm_grab_status, arm_trans_status, qda_status;
	int claw_status;
	double arm_slew_proc, arm_grab_proc, arm_trans_proc, qda_proc, cables_proc;
	double claw_proc, box_proc, clawrot1_proc, clawrot2_proc;
	double piston_rot_proc, piston_trans_proc;
	SpotLight *slca1_light;
	SpotLight *slca2_light;
	double ps_level, timer, gravity, ps_level2, ps_level3, timer2, gcadjust, ps_level4, timer3;
	double trans_adjust,att_dist, ta_mem;
	int att_dist_status;
	VECTOR3 bspos[4];
	BEACONLIGHTSPEC beac_spec[4];
	ANIMATIONCOMPONENT_HANDLE Section0, Section1, Section2, Section3, Section4, Section5;
	ATTACHMENTHANDLE ss_launch, ss_catch;
	char tgtname[256];
	double a_side_init, b_side, c_side;
	double A_angle_init, B_angle_init, C_angle_init;
	VECTOR3 qda_vector_init;
	double  piston_length_init;


};
// Define animation sequences for moving parts
// --------------------------------------------------------------
void bc_tower::DefineAnimations (void)
{
	// arm trans
	anim_arm_trans = CreateAnimation (0);
	static UINT ArmGrp4[1] = {9};
	ArmTrans = new MGROUP_TRANSLATE(0, ArmGrp4, 1, _V(0,-arm_trans_length,0));
	Section0 = AddAnimationComponent (anim_arm_trans, 0, 1, ArmTrans);

	anim_cables = CreateAnimation (0);
	static UINT ArmGrp5[1] = {4};
	ArmCables = new MGROUP_SCALE(0, ArmGrp5, 1, _V(0,143.61,0), _V(1,21.408,1));
	AddAnimationComponent (anim_cables, 0, 1, ArmCables);



	// arm slew
	anim_arm_slew = CreateAnimation (0.5);

	static UINT ArmGrp0[2] = {0,1};
	ArmSlew = new MGROUP_ROTATE(0, ArmGrp0, 0, _V(-0.024,0,-16.357),_V(0,1,0), 120*RAD);
	Section1 = AddAnimationComponent (anim_arm_slew, 0, 1, ArmSlew, Section0);


	// arm_grab
	anim_arm_grab = CreateAnimation (0);

	static UINT ArmGrp1[1] = {1};
	ArmGrab1 = new MGROUP_ROTATE(0, ArmGrp1, 1, _V(-0.024,0,-16.357),_V(0,1,0), -30*RAD);
	AddAnimationComponent (anim_arm_grab, 0, 1, ArmGrab1, Section1);

	static UINT ArmGrp2[1] = {0};
	ArmGrab2 = new MGROUP_ROTATE(0, ArmGrp2, 1, _V(-0.024,0,-16.357),_V(0,1,0), 30*RAD);
	AddAnimationComponent (anim_arm_grab, 0, 1, ArmGrab2, Section1);

		///// qda
	anim_qda = CreateAnimation (0);

	static UINT QDAGrp0[5] = {28,29,30,31,32};
	QDARot = new MGROUP_ROTATE(0, QDAGrp0, 5, _V(6.52,0,-36.883),_V(0,1,0), 60*RAD);
	Section2 = AddAnimationComponent (anim_qda, 0.025, 1, QDARot);

	static UINT QDAGrp3[2] = {2,37};
	QDATrans = new MGROUP_TRANSLATE(0, QDAGrp3, 2, _V(0,0,-1.0));
	AddAnimationComponent (anim_qda, 0, 0.025, QDATrans, Section2);

	anim_box = CreateAnimation (0);
	static UINT QDAGrp1[1] = {35};
	QDABoxRot = new MGROUP_ROTATE(0, QDAGrp1, 1, _V(0,25.116,-7.387),_V(1,0,0), -50*RAD);
	AddAnimationComponent (anim_box, 0.8, 0.9, QDABoxRot);

	static UINT QDAGrp2[3] = {33,36,44};
	QDABoxTrans = new MGROUP_TRANSLATE(0, QDAGrp2, 3, _V(0,0,1.3));
	AddAnimationComponent (anim_box, 0.9, 1, QDABoxTrans);

	////////////// claw ////
	anim_claw = CreateAnimation (0);

	static UINT ClawGrp0[1] = {23};
	ClawRetract = new MGROUP_TRANSLATE(0, ClawGrp0, 1, _V(0,0,claw_trans_length));
	Section3 = AddAnimationComponent (anim_claw, 0, 1, ClawRetract, Section2);

	anim_clawrot1 = CreateAnimation (0);
	anim_clawrot2 = CreateAnimation (0); 


	static UINT ClawGrp1[1] = {26};
	ClawRot1 = new MGROUP_ROTATE(0, ClawGrp1, 1, _V(1.984,0,-10.782),_V(0,1,0), PI2);
	AddAnimationComponent (anim_clawrot1, 0, 1, ClawRot1, Section3);

	static UINT ClawGrp2[3] = {27,38,40};
	ClawRot2 = new MGROUP_ROTATE(0, ClawGrp2, 3, _V(1.223,0,-7.571),_V(0,1,0), -PI2);
	AddAnimationComponent (anim_clawrot2, 0, 1, ClawRot2, Section2);

	static UINT ClawGrp3[1] = {24};
	ClawRot3 = new MGROUP_ROTATE(0, ClawGrp3, 1, _V(-1.984,0,-10.782),_V(0,1,0), -PI2);
	AddAnimationComponent (anim_clawrot1, 0, 1, ClawRot3, Section3);

	static UINT ClawGrp4[3] = {25,39,41};
	ClawRot4 = new MGROUP_ROTATE(0, ClawGrp4, 3, _V(-1.223,0,-7.571),_V(0,1,0), PI2);
	AddAnimationComponent (anim_clawrot2, 0, 1, ClawRot4, Section2);

	////////////// piston ///////////////////////////////

	anim_piston_rot = CreateAnimation (0);

	static UINT PistGrp0[1] = {11};
	PistonRot = new MGROUP_ROTATE(0, PistGrp0, 1, piston_hinge,_V(0,1,0), -PI2);
	Section4 = AddAnimationComponent (anim_piston_rot, 0, 1, PistonRot);

	anim_piston_trans = CreateAnimation (0);

	static UINT PistGrp1[1] = {10};
	PistonTrans = new MGROUP_TRANSLATE(0, PistGrp1, 1,_V(piston_trans_length,0,0));
	AddAnimationComponent (anim_piston_trans, 0, 1, PistonTrans, Section4);
}


// ==============================================================
// bc_tower class interface
// ==============================================================

// ==============================================================
// Overloaded callback functions
// ==============================================================

// --------------------------------------------------------------
// Set the capabilities of the vessel class
// --------------------------------------------------------------
void bc_tower::clbkSetClassCaps (FILEHANDLE cfg)
{
	strcpy(tgtname, "AUTO");
	light_status = 0;
	lox_status = 0;
	launch_status = 0;
	catch_status = 0;
	caught = 0;
	gcadjust = 0;
	ps_level = 0;
	ps_level2 = 0;
	ps_level3 = 0;
	ps_level4 = 0;
	timer = 0;
	timer2 = 0;
	timer3 = 0;
	gravity = 9.9;
	step_time = 0;

	arm_slew_status = 0;
	arm_grab_status = 0;
	arm_trans_status = 0;
	qda_status = 0;
	arm_slew_proc = 0.5;
	arm_grab_proc = 0;
	arm_trans_proc = 0;
	qda_proc = 0;
	cables_proc = 0;
	claw_status = 0;
	claw_proc = 0;
	clawrot1_proc = 0;
	clawrot2_proc = 0;
	box_proc = 1;
	piston_rot_proc = 0;
	piston_trans_proc = 0;
	piston_length_init = 0;
	qda_vector_init = _V(1,0,0);

	a_side_init = 0;
	b_side = 0;
	c_side = 0;
	A_angle_init = 0;
	B_angle_init = 0;
	C_angle_init = 0;

	trans_adjust = 0;
	att_dist = 10;
	ta_mem = 0;
	att_dist_status = 0;

	bspos[0] = _V(-9.925,145.8,-28.673);
	bspos[1] = _V(-0.051,145.8,-18.766);
	bspos[3] = _V( 0,145.8,-38.656);
	bspos[2] = _V( 9.915,145.8,-28.739);

	gcInitialize();
	DefineAnimations();

	// physical specs
	SetEmptyMass (1000000);
	SetCW (1, 1, 1, 1);
	SetSize (70);
	SetCrossSections (_V(100,100,100));
	SetPMI (_V(100,100,100));
	SetTouchdownPoints(tdvtx_geardown,ntdvtx_geardown);
	SetRotDrag (_V(1,1,1));

	SetMeshVisibilityMode ((AddMesh ("spacex_starship/bc_tower")), MESHVIS_ALWAYS);
	SetCameraOffset (_V(-5,120,-20));

	PARTICLESTREAMSPEC contrail_hover1 = {
		0, 0.5, 15, 0.5, 15, 2.0, 1.5, 0.4, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	PARTICLESTREAMSPEC contrail_hover1a = {
		0, 0.3, 20, 5, 0.25, 2.0, 0.5, 0.75, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	PARTICLESTREAMSPEC contrail_hover1b = {
		0, 0.5, 20, 8, 0.15, 1.2, 0.5, 1.25, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	///// for d3d9
	PARTICLESTREAMSPEC contrail_hover2 = {
		0, 0.6, 15, 0.5, 15, 3.0, 3, 0.4, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	PARTICLESTREAMSPEC contrail_hover2a = {
		0, 0.4, 20, 5, 0.25, 2.0, 1, 0.75, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};

	PARTICLESTREAMSPEC contrail_hover2b = {
		0, 0.6, 20, 8, 0.15, 1.2, 1, 1.25, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_FLAT, 1, 1,
	};


	if (gcEnabled() == true)
	{
		AddParticleStream (&contrail_hover2, _V(0,22,0), _V(0,-1,0), &ps_level);
		AddParticleStream (&contrail_hover2a, _V(0,26, -2.55), _V(0,0,-1), &ps_level);
		AddParticleStream (&contrail_hover2a, _V(0,90, -2.55), _V(0,0,-1), &ps_level);
	//	AddParticleStream (&contrail_hover2a, _V(-4.3,26, 1.053), _V(-0.966,0,0.259), &ps_level);
	//	AddParticleStream (&contrail_hover2a, _V(4.3,26, 1.053), _V(0.966,0,0.259), &ps_level);
		AddParticleStream (&contrail_hover2b, _V(0,60, -4.55), _V(0,0,-1), &ps_level2);
		AddParticleStream (&contrail_hover2b, _V(0.1,116.3, -4.55), _V(0,0,-1), &ps_level4);
		gcadjust = 0.005;
	}
	else 
	{
		AddParticleStream (&contrail_hover1, _V(0,22,0), _V(0,-1,0), &ps_level);
		AddParticleStream (&contrail_hover1a, _V(0,26, -2.55), _V(0,0,-1), &ps_level);
		AddParticleStream (&contrail_hover1a, _V(0,90, -2.55), _V(0,0,-1), &ps_level);
	//	AddParticleStream (&contrail_hover1a, _V(-4.3,26, 1.053), _V(-0.966,0,0.259), &ps_level);
	//	AddParticleStream (&contrail_hover1a, _V(4.3,26, 1.053), _V(0.966,0,0.259), &ps_level);
		AddParticleStream (&contrail_hover1b, _V(0,60,-4.55), _V(0,0,-1), &ps_level2);
		AddParticleStream (&contrail_hover1b, _V(0.1,116.3,-4.55), _V(0,0,-1), &ps_level4);
	}
	


	//////////////////// launch smoke ////////////////////////////////////////

	// big launch smoke
	SURFHANDLE boom1 = oapiRegisterParticleTexture ("Contrail4");

	PARTICLESTREAMSPEC contrail_steam = {
		0, 40, 15, 260, 2.6, 1.5, 24, 3, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};

	PARTICLESTREAMSPEC contrail_steam2 = {
		0, 30, 9, 260, 1.6, 0.8, 30, 3, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 4,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1,
		boom1
	};

	if (gcEnabled() == true)
	{
		contrail_steam.atmslowdown = 2.5;
		contrail_steam.lifetime = 3;
		contrail_steam2.atmslowdown = 2.5;
		contrail_steam2.lifetime = 1.6;
	}

	AddParticleStream (&contrail_steam, _V(0,-16,0), _V(0,-1,0), &ps_level3);
	AddParticleStream (&contrail_steam2, _V(0,-16,0), _V(0,-1,0), &ps_level3);

	// ***************** thruster definitions *******************

	COLOUR4 col_d = {1,1,0.9,0};
	COLOUR4 col_s = {1,1,0.9,0};
	COLOUR4 col_a = {0,0,0,0};
	COLOUR4 col_white = {1,1,1,0};
	slca1_light = (SpotLight*)AddSpotLight(_V(80,7,80), _V(-0.612,0.5,-0.612), 5000, 1e-4, 0, 1e-4, RAD*90, RAD*150, col_d, col_s, col_a);
	slca1_light->Activate(false);

	slca2_light = (SpotLight*)AddSpotLight(_V(-80,7,-80), _V(0.612,0.5,0.612), 5000, 1e-4, 0, 1e-4, RAD*90, RAD*150, col_d, col_s, col_a);
	slca2_light->Activate(false);

	ss_launch = CreateAttachment (false, _V( 0, 21.7, 0), _V( 0,1,0), _V(0,0,-1), "SS");//1.849
	ss_catch = CreateAttachment (false, _V( 0, 137.862, 0), _V( 0,1,0), _V(0,0,-1), "BST");

	/////////////////////////////
	for (int i = 0; i < 4; i++)
	{
	beac_spec[i].shape = BEACONSHAPE_STAR;
	beac_spec[i].pos = &bspos[i];
	beac_spec[i].col = &bsred;
	beac_spec[i].size = 0.5;
	beac_spec[i].falloff = 0.5;
	beac_spec[i].period = 4;
	beac_spec[i].duration = 0.2;
	beac_spec[i].tofs = 0;
	beac_spec[i].active = false;
	AddBeacon (&beac_spec[i]);
	}

}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
bc_tower::bc_tower (OBJHANDLE hVessel, int flightmodel)
: VESSEL3 (hVessel, flightmodel)
{
 
}

bc_tower::~bc_tower ()
{
	delete ArmSlew;
	delete ArmGrab1;
	delete ArmGrab2;
	delete ArmTrans;
	delete QDARot;
	delete QDATrans;
	delete ArmCables;
	delete ClawRetract;
	delete ClawRot1;
	delete ClawRot2;
	delete ClawRot3;
	delete ClawRot4;
	delete QDABoxRot;
	delete QDABoxTrans;
	delete PistonRot;
	delete PistonTrans;
}


// ==============================================================
// Overloaded clbkPostCreation callback functions
// ==============================================================

void bc_tower::clbkVisualCreated (VISHANDLE vis, int refcount)
{
	void DefineAnimations (void);
	SetAnimation(anim_arm_slew, arm_slew_proc);
	SetAnimation(anim_arm_grab, arm_grab_proc);
	SetAnimation(anim_arm_trans, arm_trans_proc);
	SetAnimation(anim_cables, cables_proc);
	SetAnimation(anim_qda, qda_proc);
	SetAnimation(anim_claw, claw_proc);
	SetAnimation(anim_clawrot1, clawrot1_proc);
	SetAnimation(anim_clawrot2, clawrot2_proc);
	SetAnimation(anim_box, box_proc);
	SetAnimation(anim_piston_rot, piston_rot_proc);
	SetAnimation(anim_piston_trans, piston_trans_proc);
}

void bc_tower::clbkPostCreation ()
{
	hGref = GetGravityRef();
	double gref_mass = oapiGetMass(hGref);
	VECTOR3 rel_pos = _V(0,0,0);
	GetRelativePos(hGref, rel_pos);
	double rad = length(rel_pos);
	gravity = GGRAV*gref_mass/(rad*rad);

	//////////// claw

	VECTOR3 vCA = vA - vC;
	VECTOR3 vAB = vB - vA;
	VECTOR3 vBC = vC - vB;

	a_side_init = length(vBC);
	b_side = length(vCA);
	c_side = length(vAB);

	double a_side_init2 = a_side_init*a_side_init;
	double b_side2 = b_side*b_side;
	double c_side2 = c_side*c_side;

	double cosaai = (b_side2 + c_side2 - a_side_init2)/(2*b_side*c_side);
	if(cosaai > 1)cosaai = 1;
	else if(cosaai < -1)cosaai = -1;
	A_angle_init = acos(cosaai);
	if(A_angle_init < 0)A_angle_init = -A_angle_init;

	double cosbai = (a_side_init2 + c_side2 - b_side2)/(2*a_side_init*c_side);
	if(cosbai > 1)cosbai = 1;
	else if(cosbai < -1)cosbai = -1;
	B_angle_init = acos(cosbai);
	if(B_angle_init < 0)B_angle_init = -B_angle_init;

	C_angle_init = PI - (A_angle_init + B_angle_init);

	/////////////////////// piston

	qda_vector_init = piston_attach - qda_rot;
	VECTOR3 piston_vector_init = piston_end - piston_hinge;
	piston_length_init = length(piston_vector_init);

}

DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new bc_tower (hvessel, flightmodel);
}
// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (bc_tower*)vessel;
}


void bc_tower::clbkSaveState (FILEHANDLE scn)
{
	char cbuf[256];

	for (int i = 0; i < (int)strlen(tgtname); i++)
	{
		if (tgtname[i] == ' ')
		{
			tgtname[i] = '+';
			break;
		}
	}

	// default vessel parameters
	VESSEL3::clbkSaveState (scn);

	sprintf (cbuf, "%d %d %d %d",  arm_slew_status, arm_grab_status, arm_trans_status, qda_status);
	oapiWriteScenario_string (scn, "ARMSTS", cbuf);

	sprintf (cbuf, "%0.3f %0.3f %0.3f %0.3f %0.3f %0.3f",  arm_slew_proc, arm_grab_proc, arm_trans_proc, cables_proc, qda_proc, box_proc);
	oapiWriteScenario_string (scn, "ARMPRC", cbuf);

	sprintf (cbuf, "%0.3f %0.3f %d %d",  att_dist, trans_adjust, caught, att_dist_status);
	oapiWriteScenario_string (scn, "PRMS", cbuf);

	sprintf (cbuf, "%d %0.3f %0.3f %0.3f",  claw_status, claw_proc, clawrot1_proc, clawrot2_proc);
	oapiWriteScenario_string (scn, "CLAW", cbuf);

	sprintf (cbuf, "%0.3f %0.3f",  piston_rot_proc, piston_trans_proc);
	oapiWriteScenario_string (scn, "PSTN", cbuf);

	// custom parameters
	sprintf (cbuf, "%d %d %d %d %0.1f",  catch_status, launch_status, light_status, lox_status, timer);
	oapiWriteScenario_string (scn, "STTS", cbuf);

	sprintf (cbuf, "%s", tgtname);
	oapiWriteScenario_string (scn, "TRGT", cbuf);

}
void bc_tower::clbkLoadStateEx (FILEHANDLE scn, void *vs)
{
    char *line;
	while (oapiReadScenario_nextline (scn, line)) {
		if (!strnicmp (line, "ARMSTS", 6)) {
			sscanf (line+6, "%d%d%d%d", &arm_slew_status, &arm_grab_status, &arm_trans_status, &qda_status);
		}else if (!strnicmp (line, "ARMPRC", 6)) {
			sscanf (line+6, "%lf%lf%lf%lf%lf%lf", &arm_slew_proc, &arm_grab_proc, &arm_trans_proc, &cables_proc, &qda_proc, &box_proc);
		}else if (!strnicmp (line, "PRMS", 4)) {
			sscanf (line+4, "%lf%lf%d%d", &att_dist, &trans_adjust, &caught, &att_dist_status);
		}else if (!strnicmp (line, "CLAW", 4)) {
			sscanf (line+4, "%d%lf%lf%lf", &claw_status, &claw_proc, &clawrot1_proc, &clawrot2_proc);
		}else if (!strnicmp (line, "PSTN", 4)) {
			sscanf (line+4, "%lf%lf", &piston_rot_proc, &piston_trans_proc);
		}else if (!strnicmp (line, "STTS", 4)) {
			sscanf (line+4, "%d%d%d%d%lf", &catch_status, &launch_status, &light_status, &lox_status, &timer);
		} else if (!strnicmp (line, "TRGT", 4)) {
			sscanf (line+4, "%s", &tgtname);
		}else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
			}
   }

	SetAnimation(anim_arm_slew, arm_slew_proc);
	SetAnimation(anim_arm_grab, arm_grab_proc);
	SetAnimation(anim_arm_trans, arm_trans_proc);
	SetAnimation(anim_cables, cables_proc);
	SetAnimation(anim_qda, qda_proc);
	SetAnimation(anim_claw, claw_proc);
	SetAnimation(anim_clawrot1, clawrot1_proc);
	SetAnimation(anim_clawrot2, clawrot2_proc);
	SetAnimation(anim_box, box_proc);
	SetAnimation(anim_piston_rot, piston_rot_proc);
	SetAnimation(anim_piston_trans, piston_trans_proc);

	for (int i = 0; i < (int)strlen(tgtname); i++)
	{
		if (tgtname[i] == '+')
		{
			tgtname[i] = ' ';
			break;
		}
	}
}


bool AddPayload(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
	OBJHANDLE hPld = oapiGetObjectByName(cbuf);
	if (hPld == NULL)return false;
	else
	{
	VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());

	ATTACHMENTHANDLE hPldAtt;
	hPldAtt = vme->GetAttachmentHandle(false, 0);
	if(vme->GetAttachmentStatus(hPldAtt) != NULL)return false;

	VESSEL2* v = (VESSEL2*)oapiGetVesselInterface(hPld);
	DWORD ac = v->AttachmentCount(true);
	if (ac == 0) return false;
	else 
	{	int i;
		for (i = 0; i < ac; i++)
		{
			ATTACHMENTHANDLE hAtt = v->GetAttachmentHandle(true, i);
			hAtt = v->GetAttachmentHandle (true, i);
			const char *id = v->GetAttachmentId (hAtt);
			if (!strnicmp (id, "LP", 2))
			{
				if(v->GetAttachmentStatus(hAtt) != NULL)return false;
				if (vme->AttachChild(hPld, hPldAtt, hAtt) == true) return true;
			}
		}
	}
	

	}

	return false;
}


bool SetCTgt(void *id, char *str, void *data)
{
	char cbuf[256];
	sscanf(str, "%s", &cbuf);
//	OBJHANDLE bbtgt = oapiGetObjectByName(cbuf);
	if (strlen(str) == 0)strcpy(cbuf, "AUTO");
//	else 
//		{
		VESSEL2* vme = (VESSEL2*)oapiGetVesselInterface(oapiGetFocusObject ());
		vme->SendBufferedKey(COMM_SEND_CT, false, cbuf);
		return true;
//		}
}


int bc_tower::clbkConsumeBufferedKey(DWORD key, bool down, char *kstate) {

	/////////// internal comms! ///////////////////
	if (key == COMM_SEND_CT) {
			sscanf (kstate, "%s", &tgtname);
			return 1;
       }
	////////////////// normal keypress ///////////////////////////

  if (!down) return 0; // only process keydown events

  if (KEYMOD_SHIFT (kstate)) {

  } else { // unmodified keys
    switch (key) {

	case OAPI_KEY_G: // move qda
		if(qda_status == 0)qda_status = 1;
		else if(qda_status == 1)qda_status = 3;
		else if(qda_status == 2)qda_status = 3;
		else if(qda_status == 3)qda_status = 1;
		return 1;


	case OAPI_KEY_B: // enable catch
		if(catch_status == 0) EnableCatch();
	//	{
	//		catch_status = 1; // enable
	//		arm_grab_status = 1;
	//	}
		else DisableCatch(); 
	//		catch_status = 0; // disable
		return 1;

	case OAPI_KEY_K: // lights on/off
		if (light_status == 0) light_status = 1;
		else if (light_status == 1) light_status = 0;
		return 1;

	case OAPI_KEY_V: // LOX vent on/off
		if (lox_status == 0)lox_status = 1;
		else lox_status = 0;
		return 1;

	case OAPI_KEY_P:
		hGref = GetGravityRef();
		oapiOpenInputBox("Enter Catch Target Name",SetCTgt, 0, 50, 0);
		return 1;

	case OAPI_KEY_N:
		oapiOpenInputBox("Attach Launcher (Enter Name)",AddPayload, 0, 50, 0);
		return 1;

	case OAPI_KEY_J: //Claw
		if (claw_status == 0)claw_status = 1;
		else if (claw_status == 1)claw_status = 3;
		else if (claw_status == 2)claw_status = 3;
		else if (claw_status == 3)claw_status = 1;
		return 1;

	case OAPI_KEY_M: //Claw remote 
		if (claw_status == 0)claw_status = 1;
		return 1;

	case OAPI_KEY_1: // ARMS trans down
		if (arm_trans_status == 0)arm_trans_status = 1;
		else if (arm_trans_status == 1)arm_trans_status = 4;
		else if (arm_trans_status == 2)arm_trans_status = 2;
		else if (arm_trans_status == 3)arm_trans_status = 1;
		else if (arm_trans_status == 4)arm_trans_status = 1;
		return 1;

	case OAPI_KEY_2: // ARMS trans up
		if (arm_trans_status == 0)arm_trans_status = 0;
		else if (arm_trans_status == 1)arm_trans_status = 3;
		else if (arm_trans_status == 2)arm_trans_status = 3;
		else if (arm_trans_status == 3)arm_trans_status = 4;
		else if (arm_trans_status == 4)arm_trans_status = 3;
		return 1;

	case OAPI_KEY_3: // ARMS open
		if (arm_grab_status == 0)arm_grab_status = 1;
		else if (arm_grab_status == 1)arm_grab_status = 4;
		else if (arm_grab_status == 2)arm_grab_status = 2;
		else if (arm_grab_status == 3)arm_grab_status = 1;
		else if (arm_grab_status == 4)arm_grab_status = 1;
		return 1;

	case OAPI_KEY_4: // ARMS close
		if (arm_grab_status == 0)arm_grab_status = 0;
		else if (arm_grab_status == 1)arm_grab_status = 3;
		else if (arm_grab_status == 2)arm_grab_status = 3;
		else if (arm_grab_status == 3)arm_grab_status = 4;
		else if (arm_grab_status == 4)arm_grab_status = 3;
		return 1;

	case OAPI_KEY_5: // ARMS slew right
		if (arm_slew_status == 0)arm_slew_status = 1;
		else if (arm_slew_status == 1)arm_slew_status = 6;
		else if (arm_slew_status == 2)arm_slew_status = 1;
		else if (arm_slew_status == 3)arm_slew_status = 1;
		else if (arm_slew_status == 4)arm_slew_status = 1;
		else if (arm_slew_status == 5)arm_slew_status = 1;
		else if (arm_slew_status == 6)arm_slew_status = 1;
		return 1;

	case OAPI_KEY_6: // ARMS slew left
		if (arm_slew_status == 0)arm_slew_status = 3;
		else if (arm_slew_status == 1)arm_slew_status = 3;
		else if (arm_slew_status == 2)arm_slew_status = 3;
		else if (arm_slew_status == 3)arm_slew_status = 6;
		else if (arm_slew_status == 4)arm_slew_status = 3;
		else if (arm_slew_status == 5)arm_slew_status = 3;
		else if (arm_slew_status == 6)arm_slew_status = 3;
		return 1;

	case OAPI_KEY_7: // ARMS slew centre
		if (arm_slew_status == 0)arm_slew_status = 5;
		else if (arm_slew_status == 1)arm_slew_status = 5;
		else if (arm_slew_status == 2)arm_slew_status = 5;
		else if (arm_slew_status == 3)arm_slew_status = 5;
		else if (arm_slew_status == 4)arm_slew_status = 5;
		else if (arm_slew_status == 5)arm_slew_status = 5;
		else if (arm_slew_status == 6)arm_slew_status = 5;
		return 1;

	case OAPI_KEY_8: // att dist further
		if (att_dist_status == 0)att_dist_status = 1;
		else if (att_dist_status == 1)att_dist_status = 0;
		else if (att_dist_status == 2)att_dist_status = 1;
		return 1;

	case OAPI_KEY_9: // att dist nearer
		if (att_dist_status == 0)att_dist_status = 2;
		else if (att_dist_status == 1)att_dist_status = 2;
		else if (att_dist_status == 2)att_dist_status = 0;
		return 1;
	}
  }
  return 0;
}
void bc_tower::clbkPreStep(double SimT, double SimDT, double MJD)
{

	step_time = SimDT;
/////////////////////// catch /////////////////////////////
	
	if(catch_status == 1)DoCatch();
	

//////////////////	Launch //////////////////////////////////
	OBJHANDLE hV = NULL;
	hV = GetAttachmentStatus(ss_launch);

	///////////// lox vents ///////////////////
	if (lox_status == 0)
	{
	if(hV == NULL) 
		{
		ps_level = 0;
		lox_status = 1;
		}
	else ps_level = 1;
	}
	else ps_level = 0;

	if(oapiGetTimeAcceleration() > 10) ps_level = 0;

	if(ps_level == 1)
	{
		if(oapiRand() > (0.99+gcadjust) && ps_level2 == 0)ps_level2 = 1;
		else if(ps_level2 == 1)
		{
			timer2 = timer2 + SimDT;
			if(timer2 > 6)
			{
			timer2 = 0;
			ps_level2 = 0;
			}
		}

		if(oapiRand() > (0.99+gcadjust) && ps_level4 == 0)ps_level4 = 1;
		else if(ps_level4 == 1)
		{
			timer3 = timer3 + SimDT;
			if(timer3 > 6)
			{
			timer3 = 0;
			ps_level4 = 0;
			}
		}
	}
	else 
	{
		ps_level2 = 0;
		ps_level4 = 0;
	}
	/////////////// launch status and detach //////////
	
	if(launch_status == 0)
	{
	if (hV != NULL) launch_status = 1;	
	}
	if(launch_status == 1)
	{
		if(hV == NULL) launch_status = 0;
		else
		{
		VESSEL *v = oapiGetVesselInterface (hV);
		VECTOR3 thrust_vector = _V(0,0,0);
		v->GetThrustVector(thrust_vector);
		double mains_level = v->GetThrusterGroupLevel(THGROUP_MAIN);
		double weight = gravity*(v->GetMass());
		double net_z_force = thrust_vector.z - weight;
			if(net_z_force > 0)
			{
			DetachChild(ss_launch, 0.0);
			ps_level = 0;
			launch_status = 2;
			}
			if(mains_level > 0)
			{
				ps_level3 = 1;
				if(qda_status == 0)qda_status = 1; 
			}
			else ps_level3 = 0;

			//if(catch_status == 0)
			//{
			//if(arm_slew_status != 4)arm_slew_status = 3;
			//if(arm_grab_status != 0)arm_grab_status = 3;
			//if(arm_trans_status != 2)arm_trans_status = 1;
			//}
		}
	}
	if(launch_status == 2) /// timing lights
		{
			timer = timer + SimDT;
			if(timer > 10)
			{
			launch_status = 0;
			light_status = 1;
			timer = 0;
			ps_level3 = 0;
			}

		}
	
	/////////////////////shadowcheck /  lights

	if(light_status == 0)
	{
	VECTOR3 rsuninv;
	VECTOR3 rsunloc;
	VECTOR3 rsun;
	GetGlobalPos(rsun);
	rsuninv = _V(-rsun.x, -rsun.y, -rsun.z);
	Global2Local(rsuninv,rsunloc);

	if (rsunloc.y < 0 && light_status == 0)
			{
			slca1_light->Activate(true);
			slca2_light->Activate(true);
			for (int i = 0; i < 5; i++)
				{
				beac_spec[i].active = true;
				}
			}
	else if (rsunloc.y > 0)
			{
			slca1_light->Activate(false);
			slca2_light->Activate(false);
			for (int i = 0; i < 5; i++)
				{
				beac_spec[i].active = false;
				}
			}
	}
	else if (light_status == 1)
	{
		slca1_light->Activate(false);
		slca2_light->Activate(false);
		for (int i = 0; i < 5; i++)
				{
				beac_spec[i].active = false;
				}
	}
	
	//////////////////////////////// ANIMATIONS
	//// qda
	if(qda_status == 0) qda_proc = 0;
	else if (qda_status == 1)
		{
			double da = SimDT*QDA_DETACH_SPEED;
			qda_proc = qda_proc+da;
			if(qda_proc > 1)
			{
			qda_proc = 1;
			qda_status = 2;
			}
		}
	else if (qda_status == 2)qda_proc = 1;
	else if (qda_status == 3)
		{
			double da = SimDT*QDA_ATTACH_SPEED;
			qda_proc = qda_proc-da;
			if(qda_proc < 0)
			{
			qda_proc = 0;
			qda_status = 0;
			}
		}
	SetAnimation(anim_qda, qda_proc);
	box_proc = 1 - qda_proc;
	SetAnimation(anim_box, box_proc);


/////////////// claw /////////////////////////////////////
	if(claw_status == 0) claw_proc = 0;
	else if (claw_status == 1)
		{
			double da = SimDT*CLAW_SPEED;
			claw_proc = claw_proc+da;
			if(claw_proc > 1)
			{
			claw_proc = 1;
			claw_status = 2;
			}
		}
	else if (claw_status == 2)claw_proc = 1;
	else if (claw_status == 3)
		{
			double da = SimDT*CLAW_SPEED;
			claw_proc = claw_proc-da;
			if(claw_proc < 0)
			{
			claw_proc = 0;
			claw_status = 0;
			}
		}
	SetAnimation(anim_claw, claw_proc);

	VECTOR3 new_vC = vC + _V(0,0,(claw_proc*claw_trans_length));
	VECTOR3 new_vCB = vB - new_vC;
	VECTOR3 vCB_init = vB - vC;

	VECTOR3 new_vCB_norm = new_vCB;
	normalise(new_vCB_norm);
	VECTOR3 vCB_init_norm = vCB_init;
	normalise(vCB_init_norm);

	double cosfra = dotp(new_vCB_norm,vCB_init_norm);
	if(cosfra > 1)cosfra = 1;
	else if(cosfra < -1)cosfra = -1;
	double frame_rot_angle = acos(cosfra);

	double a_side = length(new_vCB);
	double a_side2 = a_side*a_side;
	double b_side2 = b_side*b_side;
	double c_side2 = c_side*c_side;

	double cosA = ((b_side2 + c_side2 - a_side2)/(2*b_side*c_side));
	if(cosA > 1) cosA = 1;
	else if(cosA < -1) cosA = -1;

	double cosB = ((a_side2 + c_side2 - b_side2)/(2*a_side*c_side));
	if(cosB > 1) cosB = 1;
	else if(cosB < -1) cosB = -1;

	double A_angle = acos(cosA);
	if(A_angle < 0)A_angle = A_angle + PI2;
	double B_angle = acos(cosB);
	if(B_angle < 0)B_angle = B_angle + PI2;

	double C_angle = PI - (A_angle + B_angle);

	double B_req = B_angle - B_angle_init - frame_rot_angle;
	if(B_req < 0)B_req = B_req + PI2;
	else if(B_req > PI2)B_req = B_req - PI2;

	double C_req = C_angle - C_angle_init + frame_rot_angle;
	if(C_req < 0)C_req = C_req + PI2;
	else if(C_req > PI2)C_req = C_req - PI2;

	clawrot1_proc = C_req/PI2;
	clawrot2_proc = B_req/PI2;

	SetAnimation(anim_clawrot1, clawrot1_proc);
	SetAnimation(anim_clawrot2, clawrot2_proc);


///////// arm slew /////////////////////////////////
if(catch_status == 0) // disabled during catch
	{
	if(arm_slew_status == 0) arm_slew_proc = 0.5; // 0 centred
	else if(arm_slew_status == 1)					// 1 go right
		{
			double da = SimDT*ARM_SLEW_SPEED;
			arm_slew_proc = arm_slew_proc+da;
			if(arm_slew_proc > 1)
			{
				arm_slew_proc = 1;
				arm_slew_status = 2;
			}
		}
	else if(arm_slew_status == 2) arm_slew_proc = 1;// 2 full right
	else if(arm_slew_status == 3)					// 3 go left
		{
			double da = SimDT*ARM_SLEW_SPEED;
			arm_slew_proc = arm_slew_proc-da;
			if(arm_slew_proc < 0)
			{
				arm_slew_proc = 0;
				arm_slew_status = 4;
			}
		}
	else if(arm_slew_status == 4) arm_slew_proc = 0; //4 full left
	else if(arm_slew_status == 5)					// 5 return to centre
		{
			int slew_dir = 1;
			if(arm_slew_proc > 0.5)slew_dir = -1;
			double da = SimDT*ARM_SLEW_SPEED*slew_dir;

			arm_slew_proc = arm_slew_proc+da;
			if(slew_dir == 1)
			{
			if(arm_slew_proc > 0.5)
				{
				arm_slew_proc = 0.5;
				arm_slew_status = 0;
				}
			}
			else
			{
			if(arm_slew_proc < 0.5)
				{
				arm_slew_proc = 0.5;
				arm_slew_status = 0;
				}
			}
		}
	else if(arm_slew_status == 6)arm_slew_proc = arm_slew_proc;
	SetAnimation(anim_arm_slew, arm_slew_proc);

	
	////////////////////////////////////////
	//// up down trans
	if(arm_trans_status == 0)arm_trans_proc = 0; // full up
	else if(arm_trans_status == 1)// lowering
			{
				double da = step_time*ARM_TRANS_SPEED*0.2;
				arm_trans_proc = arm_trans_proc+da;
				if(arm_trans_proc > 1)
				{
					arm_trans_proc = 1;
					arm_trans_status = 2;
				}
			}
	else if(arm_trans_status == 2)arm_trans_proc = 1; // full down
	else if(arm_trans_status == 3)// raising
			{
				double da = step_time*ARM_TRANS_SPEED*0.2;
				arm_trans_proc = arm_trans_proc-da;
				if(arm_trans_proc < 0)
				{
					arm_trans_proc = 0;
					arm_trans_status = 0;
				}
			}
	else if(arm_trans_status == 4)arm_trans_proc = arm_trans_proc; // stopped

			SetAnimation(anim_arm_trans, arm_trans_proc);
			SetAnimation(anim_cables, arm_trans_proc);

	
	

	} // disabled during catch

/////////////// att dist /////////////////////////
	if(caught == 1)
	{
			if(att_dist_status == 1)
			{
				att_dist = att_dist + (step_time * 0.5);
				if(att_dist > 28)
				{
					att_dist = 28;
					att_dist_status = 0;
				}
			}
			else if(att_dist_status == 2)
			{
				att_dist = att_dist - (step_time * 0.5);
				if(att_dist < 8)
				{
					att_dist = 8;
					att_dist_status = 0;
				}
			}
	}

/////////////// set attach /////////////////////////
	double arm_slew_angle = (arm_slew_proc * (120*RAD)) - (60*RAD);
			double att_x_pos = (att_dist * (sin(arm_slew_angle))) - 0.024;
			double att_z_pos = (att_dist * (cos(arm_slew_angle))) - 16.357;
			double att_y_pos = 137.862 - (arm_trans_length*arm_trans_proc);
			VECTOR3 new_att_pos = _V(att_x_pos, att_y_pos, att_z_pos);
			VECTOR3 new_att_rot = _V(-sin(arm_slew_angle),0,-cos(arm_slew_angle));
			SetAttachmentParams(ss_catch,new_att_pos,_V(0,1,0),new_att_rot);

	//////////// arm grab ///////////////////////////////

	if(arm_grab_status == 0) arm_grab_proc = 0;
	else if (arm_grab_status == 1)
		{
			double da = SimDT*ARM_GRAB_SPEED;
			arm_grab_proc = arm_grab_proc+da;
			if(arm_grab_proc > 1)
			{
			arm_grab_proc = 1;
			arm_grab_status = 2;
			}
		}
	else if (arm_grab_status == 2)arm_grab_proc = 1;
	else if (arm_grab_status == 3)
		{
			double da = SimDT*ARM_GRAB_SPEED;
			arm_grab_proc = arm_grab_proc-da;
			if(arm_grab_proc < 0)
			{
			arm_grab_proc = 0;
			arm_grab_status = 0;
			}
		}
	else if (arm_grab_status == 4)arm_grab_proc = arm_grab_proc;

	SetAnimation(anim_arm_grab, arm_grab_proc);

	//////////////////////////////////////////////////////////////////////////////
	//sprintf(oapiDebugString(), "launch_status %d ", launch_status);
	if(caught == 1) // drop booster
	{
			OBJHANDLE hPadVes = GetAttachmentStatus(ss_launch);
			OBJHANDLE hBooster = GetAttachmentStatus(ss_catch);
		if(hPadVes == NULL)
		{
			if(arm_grab_proc > 0.3)
			{
			DetachChild(ss_catch, 0);
			caught = 0;
			
			VECTOR3 att_pos_loc = _V( 0, 21.7, 0);
			VECTOR3 att_pos_glob = _V(0,0,0);
			Local2Global(att_pos_loc, att_pos_glob);

			VESSEL *v = oapiGetVesselInterface (hBooster);
			VECTOR3 booster_att_pos_glob = _V(0,0,0);
			ATTACHMENTHANDLE hAtt;

			DWORD nAttach = v->AttachmentCount(true);
			for (DWORD i = 0; i < nAttach; i++) 
				{ // now scan all attachment points of the candidate			
				hAtt = v->GetAttachmentHandle (true, i);
				const char *id = v->GetAttachmentId (hAtt);
				if (!strnicmp (id, "LP", 2))
					{
					VECTOR3 attpos, attdir, attrot;
					v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
					v->Local2Global(attpos, booster_att_pos_glob);
					//sprintf(oapiDebugString(), "attpos x %0.3f y %0.3f z %0.3f", attpos.x,attpos.y,attpos.z);
					}
				}
			VECTOR3 att_distv = booster_att_pos_glob - att_pos_glob;
			
			//sprintf(oapiDebugString(), "att_dist %0.3f ", length(att_distv));
			if(length(att_distv) < 2.0)
				{
				AttachChild(hBooster, ss_launch, hAtt);
				
				}

			}
		}
		//else ///// already booster on pad
		//{
		//if(arm_grab_proc > 0.3)
		//	{
		//	DetachChild(ss_catch, 0);
		//	caught = 0;
		//	
		//	VECTOR3 att_pos_loc = _V( 0, (21.7+66), 0);
		//	VECTOR3 att_pos_glob = _V(0,0,0);
		//	Local2Global(att_pos_loc, att_pos_glob);

		//	VESSEL *v = oapiGetVesselInterface (hBooster);
		//	VECTOR3 booster_att_pos_glob = _V(0,0,0);
		//	ATTACHMENTHANDLE hAtt;

		//	DWORD nAttach = v->AttachmentCount(true);
		//	for (DWORD i = 0; i < nAttach; i++) 
		//		{ // now scan all attachment points of the candidate			
		//		hAtt = v->GetAttachmentHandle (true, i);
		//		const char *id = v->GetAttachmentId (hAtt);
		//		if (!strnicmp (id, "LP", 2))
		//			{
		//			VECTOR3 attpos, attdir, attrot;
		//			v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
		//			v->Local2Global(attpos, booster_att_pos_glob);
		//			//sprintf(oapiDebugString(), "attpos x %0.3f y %0.3f z %0.3f", attpos.x,attpos.y,attpos.z);
		//			}
		//		}
		//	VECTOR3 att_distv = booster_att_pos_glob - att_pos_glob;
		//	
		//	//sprintf(oapiDebugString(), "att_dist %0.3f ", length(att_distv));
		//	if(length(att_distv) < 2.0)
		//		{
		//		//launch_status = 0;
		//		DetachChild(ss_launch);
		//		oapiDeleteVessel(hPadVes);
		//		char vesclassbuf [256];
		//		char vesnamebuf [256];
		//		strcpy(vesclassbuf, v->GetClassName());
		//		strcpy(vesnamebuf, v->GetName());
		//		VESSELSTATUS vs;
		//		GetStatus(vs);
		//		oapiDeleteVessel(hBooster);
		//		OBJHANDLE hNewVes = oapiCreateVessel(vesnamebuf, vesclassbuf, vs);

		//		VESSEL *nv = oapiGetVesselInterface (hNewVes);
		//		FILEHANDLE cfg = oapiOpenFile(vesclassbuf,FILE_IN, CONFIG); 
		//		((VESSEL3*)nv)->clbkSetClassCaps(cfg);

		//		DWORD nAttach = nv->AttachmentCount(true);
		//		for (DWORD i = 0; i < nAttach; i++) 
		//			{ // now scan all attachment points of the candidate			
		//			hAtt = nv->GetAttachmentHandle (true, i);
		//			const char *id = nv->GetAttachmentId (hAtt);
		//			if (!strnicmp (id, "LP", 2))
		//				{
		//				AttachChild(hNewVes, ss_launch, hAtt);
		//				lox_status = 0;
		//				}
		//			}
		//		
		//		}

		//	}
		//}
	}

	////////////////////////////// piston ///////////////////
	MATRIX3 rm;
	double rot_angle = qda_proc*(60*RAD);
	VECTOR3 rot_axis = _V(0,1,0);
	rm = rotm(rot_axis, rot_angle);
	VECTOR3 qda_vector_new = mul(rm, qda_vector_init);
	VECTOR3 piston_attach_new = qda_vector_new + qda_rot;

	VECTOR3 hinge_attach_vector = piston_attach_new - piston_hinge;
	double piston_length_new = length(hinge_attach_vector);
	double delta_length = piston_length_new - piston_length_init;
	piston_trans_proc = delta_length/piston_trans_length;
	if(piston_trans_proc < 0)piston_trans_proc = 0;
	else if(piston_trans_proc > 1)piston_trans_proc = 1;

	normalise(hinge_attach_vector);
	double piston_rot_angle_req = asin(hinge_attach_vector.z);
	if(piston_rot_angle_req < 0)piston_rot_angle_req = piston_rot_angle_req + PI2;
	piston_rot_proc = piston_rot_angle_req/PI2;
	if(piston_rot_proc > 1)piston_rot_proc = 1;
	else if(piston_rot_proc < 0)piston_rot_proc = 0;

	SetAnimation(anim_piston_rot, piston_rot_proc);
	SetAnimation(anim_piston_trans, piston_trans_proc);
	/////////////////////////////////////////////////////////


	}

//void bc_tower::clbkPostStep (double simt, double simdt, double mjd)
//{
//	
//	
//}

	bool bc_tower::clbkDrawHUD (int mode, const HUDPAINTSPEC *hps, oapi::Sketchpad *skp)
{

	int s = hps->H;
	double d = (s*0.00130208);

	VESSEL3::clbkDrawHUD (mode, hps, skp);

	int sw = ((hps->W));
	int lw =  (int)(16*sw/1024);
	int lwoffset = sw - (18*lw);
	int hlw = (int) (lw/2);


	int roxl = 0;
	int royl = 0;

	double ds = s;
	double dsw = sw;
	double sc_ratio = ds/dsw;


	if (sc_ratio < 0.7284)
	{
	roxl = (lw*10);
	royl = (int) (-88*d);
	}

	int wd = (int)(152*d);
	int wc = (int)(168*d);
	int w0 = (int)(184*d);
	int w1 = (int) (200*d);
	int w2 = (int)(216*d);
	int w3 = (int)(232*d);
	int w4 = (int)(248*d);
	int w5 = (int)(264*d);
	int w6 = (int)(280*d);
	int w7 = (int)(296*d);
	int w8 = (int)(312*d);
//	int w9 = (int)(328*d);
//	int w10 = (int)(344*d);
//	int w11 = (int)(360*d);
//	int w12 = (int)(376*d);
//	int w13 = (int)(392*d);
//	int w14 = (int)(408*d);

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(claw_proc == 0)sprintf(abuf,"J = Engage/Disengage CLAW: ENGAGED");
	else sprintf(abuf,"J = Engage/Disengage CLAW: DISENGAGED");
	skp->Text ((10+roxl), (wd+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(qda_proc == 0)sprintf(abuf,"G = Engage/Disengage QDA: ENGAGED");
	else sprintf(abuf,"G = Engage/Disengage QDA: DISENGAGED");
	skp->Text ((10+roxl), (wc+royl), abuf,strlen(abuf));
	}

	

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(catch_status == 0)sprintf(abuf,"B = Enable Catch: OFF");
	else sprintf(abuf,"B = Enable Catch: ON");
	skp->Text ((10+roxl), (w0+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"P = Enter Catch Tgt.Name: %s", tgtname);
	skp->Text ((10+roxl), (w1+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(light_status == 0)sprintf(abuf,"K = Night Lights On/Off: ON");
	else sprintf(abuf,"K = Night Lights On/Off: OFF");
	skp->Text ((10+roxl), (w2+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	if(lox_status == 0)sprintf(abuf,"V = LOX Venting On/Off: ON");
	else sprintf(abuf,"V = LOX Venting On/Off: OFF");
	skp->Text ((10+roxl), (w3+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"N = Open Launcher Attachment Dialogue Box");
	skp->Text ((10+roxl), (w4+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"1/2 = Lower/Raise/Pause Catch Arms");
	skp->Text ((10+roxl), (w5+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"3/4 = Open/Close/Pause Catch Arms");
	skp->Text ((10+roxl), (w6+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"5/6/7 = Slew Right/Left/Centre Catch Arms");
	skp->Text ((10+roxl), (w7+royl), abuf,strlen(abuf));
	}

	{
	skp->SetTextColor(0x00FFFFFF);
	char abuf[256];
	sprintf(abuf,"8/9 = Transport Along Arms +/-");
	skp->Text ((10+roxl), (w8+royl), abuf,strlen(abuf));
	}

	return true;
	}

void bc_tower::DoCatch(void)
{
bool tgt_valid = false;
OBJHANDLE hTarget = oapiGetObjectByName(tgtname);
		VECTOR3 target_pos = _V(0,0,0);
		double target_spd = 0;
		ATTACHMENTHANDLE hAtt = NULL;

	if(hTarget == NULL)
	{
		DWORD num_ves = oapiGetVesselCount();
		for (DWORD j = 0; j < num_ves; j++)
		{
			OBJHANDLE hVes = oapiGetVesselByIndex(j);
			if(hVes == GetHandle())continue; // not myself
			if(hVes == GetAttachmentStatus(ss_launch))continue;
			VECTOR3 hvespos;
			GetRelativePos(hVes,hvespos);
			if(length(hvespos) > 1000)continue; // out of range

			VESSEL *v = oapiGetVesselInterface (hVes);
			DWORD nAttach = v->AttachmentCount(true);
			for (DWORD i = 0; i < nAttach; i++) 
			{ // now scan all attachment points of the candidate			
			hAtt = v->GetAttachmentHandle (true, i);
			const char *id = v->GetAttachmentId (hAtt);
			if (!strnicmp (id, "ARM", 3))
				{
				VECTOR3 attpos, attdir, attrot, attposglobal, target_vel;
				v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
				v->Local2Global(attpos, attposglobal);
				Global2Local(attposglobal, target_pos);
				v->GetAirspeedVector(FRAME_HORIZON,target_vel);
				target_spd = -target_vel.y;
				hTarget = hVes;
				//oapiGetObjectName(hTarget, tgtname, 50);
				tgt_valid = true;
				break;
				}
			}
			if(hTarget != NULL)break;
			

		}
	}
	else
	{
	VECTOR3 hvespos;
	GetRelativePos(hTarget,hvespos);
	if(length(hvespos) < 1000)
		{
		VESSEL *v = oapiGetVesselInterface (hTarget);
			DWORD nAttach = v->AttachmentCount(true);
			for (DWORD i = 0; i < nAttach; i++) 
			{ // now scan all attachment points of the candidate			
			hAtt = v->GetAttachmentHandle (true, i);
			const char *id = v->GetAttachmentId (hAtt);
			if (!strnicmp (id, "ARM", 3))
				{
				VECTOR3 attpos, attdir, attrot, attposglobal, target_vel;
				v->GetAttachmentParams(hAtt, attpos, attdir, attrot);
				v->Local2Global(attpos, attposglobal);
				Global2Local(attposglobal, target_pos);
				v->GetAirspeedVector(FRAME_HORIZON,target_vel);
				target_spd = -target_vel.y;
				tgt_valid = true;
				break;
				}
			
			}

		}
	else tgt_valid = false;

	}

	if(target_pos.y < 60)tgt_valid = false;
	if(target_spd < 0)tgt_valid = false;

	if(tgt_valid == true)	
	{
		double pivot_y_pos = 137.862 - (arm_trans_length*arm_trans_proc);

		VECTOR3 pivot_target_pos = target_pos - _V(-0.024,pivot_y_pos,-16.357);
			double z_angle = atan2(pivot_target_pos.x, pivot_target_pos.z);
			if(z_angle > PI)z_angle = z_angle-PI2;
			double slew_req = (z_angle/(120*RAD)) + 0.5;
			if(slew_req > 1)slew_req = 1;
			else if(slew_req < 0)slew_req = 0;
			if(arm_slew_proc > slew_req)
			{
				double da = step_time*ARM_SLEW_SPEED;
				arm_slew_proc = arm_slew_proc-da;
				if(arm_slew_proc < slew_req)arm_slew_proc = slew_req;
			}
			else if(arm_slew_proc < slew_req)
			{
				double da = step_time*ARM_SLEW_SPEED;
				arm_slew_proc = arm_slew_proc+da;
				if(arm_slew_proc > slew_req)arm_slew_proc = slew_req;
			}
			SetAnimation(anim_arm_slew, arm_slew_proc);

			double z_dist = length(_V(pivot_target_pos.x, 0,pivot_target_pos.z));

			att_dist = z_dist;
			if(att_dist < 8)att_dist = 8;
			else if(att_dist > 28)att_dist = 28;
			double arm_slew_angle = (arm_slew_proc * (120*RAD)) - (60*RAD);
			double att_x_pos = (att_dist * (sin(arm_slew_angle))) - 0.024;
			double att_z_pos = (att_dist * (cos(arm_slew_angle))) - 16.357;
			double att_y_pos = pivot_y_pos;
			VECTOR3 new_att_pos = _V(att_x_pos, att_y_pos, att_z_pos);
			VECTOR3 new_att_rot = _V(-sin(arm_slew_angle),0,-cos(arm_slew_angle));
			//SetAttachmentParams(ss_catch,new_att_pos,_V(0,1,0),new_att_rot);
			VECTOR3 tgt_att_vect = target_pos - new_att_pos;
			if (length(tgt_att_vect) < 0.3)
			{
			if(caught == 0)AttachChild(hTarget, ss_catch, hAtt);
			caught = 1;
			}

			if(caught == 0)
			{
			trans_adjust = trans_adjust + (step_time*0.1);
			if(trans_adjust > (target_spd - 1)/arm_trans_length)trans_adjust = (target_spd - 1)/arm_trans_length;//4
			if(trans_adjust < 0) trans_adjust = 0;
			ta_mem = trans_adjust;
			}
			else
			{
			trans_adjust = trans_adjust - (step_time*0.25*ta_mem);
			if(trans_adjust < 0)
				{
					trans_adjust = 0;
					catch_status = 2;
					ta_mem = 0;
				}
			}


			if(pivot_target_pos.y < 4)//16
			{
				
				double da = step_time;//*ARM_TRANS_SPEED;
				arm_trans_proc = arm_trans_proc+(da*trans_adjust);
				if(arm_trans_proc > 1)arm_trans_proc = 1;
				else if(arm_trans_proc < 0)arm_trans_proc = 0;
			}
			SetAnimation(anim_arm_trans, arm_trans_proc);
			SetAnimation(anim_cables, arm_trans_proc);

			if(pivot_target_pos.y < 10 && caught == 0)arm_grab_status = 3; // 9

			if(catch_status == 2)
				{
				trans_adjust = 0;// reset
				arm_trans_status = 4; // stopped
				arm_slew_status = 6; // stopped
				catch_status = 0;
				}
	}
	else if(caught == 0)// go to default pos
	{
			if(arm_trans_proc > 0)
			{

				double da = step_time*ARM_TRANS_SPEED;
				arm_trans_proc = arm_trans_proc-da;
				if(arm_trans_proc < 0)arm_trans_proc = 0;
			}
			SetAnimation(anim_arm_trans, arm_trans_proc);
			SetAnimation(anim_cables, arm_trans_proc);

			//////////////////////////////////////////////////////

			int slew_dir = 1;
			if(arm_slew_proc > 0.5)slew_dir = -1;
			double da = step_time*ARM_SLEW_SPEED*slew_dir;

			arm_slew_proc = arm_slew_proc+da;
			if(slew_dir == 1)
			{
			if(arm_slew_proc > 0.5)
				{
				arm_slew_proc = 0.5;
				arm_slew_status = 0;
				}
			}
			else
			{
			if(arm_slew_proc < 0.5)
				{
				arm_slew_proc = 0.5;
				arm_slew_status = 0;
				}
			}
			SetAnimation(anim_arm_slew, arm_slew_proc);

			arm_grab_status = 1;
		}
}

void bc_tower::EnableCatch(void)
{
	OBJHANDLE hVes = GetAttachmentStatus(ss_catch);
	if(hVes != NULL) oapiDeleteVessel(hVes);

	arm_trans_status = 3;
	arm_grab_status = 1;
	arm_slew_status = 5;
	qda_status = 1;
	caught = 0;
	catch_status = 1;
}

void bc_tower::DisableCatch(void)
{
	OBJHANDLE hVes = GetAttachmentStatus(ss_catch);
	if(hVes != NULL) oapiDeleteVessel(hVes);

	//arm_trans_status = 3;
	//arm_grab_status = 1;
	//arm_slew_status = 5;
	caught = 0;
	catch_status = 0;
}